
/* LINK LAYER WRAPPER ROUTINES */

extern Log pkt_log;
extern int doXstuff;

static caddr_t cn_reset(),cn_create(),cn_neighbor(),
               cn_receive(), cn_start(),
               cn_uneighbor(),cn_route(),cn_hops(),cn_send();
static caddr_t cn_carrier_sense(), cn_check_transmission_status(),
     cn_collision_detect_notify(), cn_line_may_be_clear(), cn_collision_clear();
static caddr_t  cn_line_clear_aftran(), cn_line_clear_afcoll(), cn_stop_graphic(), cn_update_interface();

/*
static void  trace();
*/

static int update_bus_info();

#define link_other_event_handlers()			\
{							\
  case EV_RESET:					\
       result=(caddr_t)cn_reset(cn);			\
       break;						\
  case EV_CREATE:					\
       result = cn_create((char *)arg);			\
       break; 						\
  case EV_DEL:						\
       comp_delete((Component *)cn);			\
       result = (caddr_t)cn;				\
       break;						\
  case EV_NEIGHBOR:					\
       result = cn_neighbor(cn, (Component *)arg);	\
       break;						\
  case EV_UNEIGHBOR:					\
       result = cn_uneighbor(cn, (Component *)arg);	\
       break;						\
  case EV_LEGAL_NEXT_HOPS:				\
       result = cn_hops(cn);				\
       break;						\
  case EV_MAKE_ROUTE:					\
       result = cn_route(cn, (list *)arg);		\
       break;						\
  case EV_START: 					\
       result = cn_start(cn);				\
       break;						\
  case EV_UPDATE_ROUTE_ACTIVITY:			\
       update_routes(src,(Component *)cn);		\
       result=(caddr_t)cn;				\
       break;                                           \
  case INFORM_MAC:					\
       update_bus_info(cn,(Component *)src,(tick_t)arg);    \
       result=(caddr_t)cn;				\
       break;						\
}

#define LINK_ENTITY_TYPE LINKConnectiont

/****************************************/
static int
update_bus_info(cn,src,d)
     register LINKConnectiont *cn;
     register Component *src;
     register tick_t d;
{
  int  i,j;
  tick_t a,b;
  long c,e;
  
     for (i=0;i<MAXNODES;i++)
	  if ( cn->node[i].c == src) break;

     if(i==MAXNODES) return -1;

     if (cn->Propogation_Time[0][i] == INFINITY)
	  ev_enqueue(INFORM_MAC, (Component *)src,
		     (Component *)cn,
		     ev_now()+1, cn->cn_action, (PDU_TYPE *) NULL,
		     (caddr_t) d);

     a = cn->Propogation_Time[0][i] =
	 cn->Propogation_Time[i][0] = (tick_t)d; /* in units of ticks */
     
     for (j=0;j<MAXNODES;j++){
	  b = cn->Propogation_Time[0][j];
	  if(b  != (tick_t)INFINITY){
	       c = a - b;
	       e = b - a;
	       cn->Propogation_Time[i][j] = Maxm(c,e);
	       cn->Propogation_Time[j][i] = cn->Propogation_Time[i][j];
	  }
     }
     
     return i;
}
/****************************************/
static caddr_t
cn_create(name)
     register char *name;
{
    LINKConnectiont *newcn;
    int i;
    
   /* Memory for the component structure. */
   newcn = (LINKConnectiont *)sim_malloc(sizeof(LINKConnectiont));

   /* First things first-- copy name into the new structure. */
   strncpy(newcn->cn_name, name, 40);

   /* have to create a neighbor list */
  newcn->cn_neighbors= l_create();
  newcn->cn_params = q_create();
  newcn->cn_type = LINK_TYPE;
  newcn->cn_class = LINK_CLASS;
  newcn->cn_action = (PFP)link_action;
  newcn->cn_menu_up = FALSE;
  newcn->cn_flash = 0;

 /* Initialize the parameters */
  newcn->cn_Name = param_init ((Component*)newcn, "Name",
             (PFD)NULL, (PFP)make_name_text, (PFP)make_short_name_text,
             param_input_name,
             0, DisplayMask | InputMask | CanHaveLogMask, 0.0);

 newcn->cn_delay = param_init((Component *)newcn,
            "Delay (usecs)",
             (PFD)NULL,(PFP)make_double_text,(PFP)make_short_double_text,
             param_input_double,
             0,DisplayMask | InputMask, 0.0);
 newcn->cn_delay->u.d = 5.0 ; /* 1km = 5usecs */

 newcn->cn_err = param_init((Component *)newcn,
            "Error Probability",
             (PFD)NULL,(PFP)make_double_text,(PFP)make_short_double_text,
             param_input_double,
             0,DisplayMask | InputMask, 0.0);
 newcn->cn_err->u.d = 0.0;

 /* $Revision$: Enet */
 newcn->num_nodes = 0; /* node 0 is the terminal point of the bus */
 for (i=0;i<MAXNODES;i++) newcn->node[i].c = (Component *)NULL;
 
   return((caddr_t)newcn);
}

/****************************************/
static caddr_t
cn_reset(cn)
     register LINKConnectiont *cn;
{
     int  i,j;

 /* $Revision$: Enet */
     for (i=0;i<MAXNODES;i++)
	  for (j=0;j<MAXNODES;j++)
	       cn->Propogation_Time[i][j] = (tick_t)INFINITY;
     
     return ((caddr_t)cn);
}

/****************************************/
static caddr_t
cn_start(cn)
     register LINKConnectiont *cn;
{
     int  i;
     Component *c;
     
 /* $Revision$: Testing deterministic drop policy */
     if (cn->cn_err->u.d > 0){
	  cn->cn_one_by_err = 1.0/cn->cn_err->u.d;
	  cn->cn_accts = cn->cn_one_by_err;
     }

     cn->Collision_Clear_Time = (tick_t)INFINITY;
     cn->Collision_On_Link = FALSE;

     /* $Revision$ : Enet variables initialization */
     for (i=0;i<=cn->num_nodes;i++){
	  cn->node[i].Signal_Arrival_Time = (tick_t)INFINITY;
	  cn->node[i].Tmp_Signal_Arrival_Time = INFINITY;
	  cn->node[i].Waiting = FALSE;
	  cn->node[i].Node_Transmitting = FALSE;
	  cn->node[i].Collision_Detected = FALSE;
	  cn->node[i].Collision_Detection_Time = (tick_t)INFINITY;
	  cn->node[i].Last_Collision_Detect_Time = (tick_t)INFINITY;
	  cn->node[i].EndTx = (tick_t)INFINITY;
	  c = cn->node[i].c;

	  if (c != (Component *) NULL)
	       ev_enqueue(INFORM_MAC, (Component *)cn,
			  (Component *)c,
			  ev_now()+1, c->co_action, (PDU_TYPE *) NULL,
			  (caddr_t) i);

     }



     return ((caddr_t)cn);
}
/****************************************** */

static int
get_number_from_string(p)
char *p;
{

while(!isdigit(*p)) p++;

return atoi(p);
}
/****************************************/
static caddr_t
cn_neighbor(cn,c)
     register LINKConnectiont *cn;
     register Component *c;
{
     register Neighbor *n;
     int i;
     
  /* $Revision$: Enet: Infinite number of neighbors possible for the Enet Bus */
    if (!(n=add_neighbor((Component*)cn,c,0,1,PHY_CLASS)))
        return ((caddr_t)NULL);
    cn->num_nodes++;
    cn->node[cn->num_nodes].c = c;
    i = get_number_from_string(c->co_name);
    cn->findindex[i] = cn->num_nodes; /* $Revision$: Assumption: Every phy is a
					 string+ number: eg: phy1 */
    
   
 
    return ((caddr_t)n);
}

/*****************************************/
static caddr_t
cn_uneighbor(cn,c)
   register LINKConnectiont *cn;
   register Component *c;
{
   return ((caddr_t)remove_neighbor((Component *)cn,c));
}

/******************************************/

static caddr_t
cn_hops(cn,l)
   register LINKConnectiont *cn;
   list *l;
{


    register list *nlist;
    register Neighbor *n;
    register Component *c;

/* what is legal for the next hop depends on the type of the previous
   hop. If the previous one is a link, then the next must be a connection.
   If the previous one is a connection, the next one may be either a link
   or a connection.
   (Allowing a connection with both ends at the same bte ) */

   /* The previous step in the route = tail of list is put in c */
   c = (Component *)(l->l_tail->le_data);

   /* Make an empty list */
   nlist = l_create();

   /* Put appropriate neighbors into list */
   for (n=(Neighbor *)cn->cn_neighbors->l_head;n;n=n->n_next)
       if (n->n_c!=c && ((c->co_class == PHY_CLASS) ||
           (c->co_class == PHY_CLASS && n->n_c->co_class == PHY_CLASS)))
            l_addt(nlist,(caddr_t)n->n_c);


    return ((caddr_t)nlist);


}

/*******************************************/
static caddr_t
cn_route(cn,l)
         register LINKConnectiont *cn;
         list *l;
{

      /* $Revision$: Enet : This is a dummy routine */

#ifdef NOTDEF
        l_elt * le;
	CRT newroute;
	char dstr[10];


	le  = l->l_head->le_next;   /* the source link layer */ 

	le = le->le_next;          /* the link */
	strcpy(newroute.link,le->le_data);
	le = le->le_next;          /* the dest link layer */
	sscanf(le->le_data,"%s%d",dstr,newroute.dest);   /* extract the num */

	l_addt(cn->cn_crt, (caddr_t) &newroute);



         Param *p;
         Component *this_component;
         Component *other_component;
         Neighbor *n1;
         Neighbor *n2;
    

         p=param_init((Component *)cn,"Other End Connection :               ",
                      int_calc,make_int_text,make_short_int_text,
                      (PFI)NULL,0,DisplayMask | CanHaveLogMask,0.0);

        if ((((Component *)route_list->l_head->le_data)->co_name[3] !=
              cn->cn_name[3]) || 
             (((Component *)route_list->l_head->le_data)->co_name[2] !=
              cn->cn_name[2]) || 
              (((Component *)route_list->l_head->le_data)->co_name[1] !=
              cn->cn_name[1]) || 
              (((Component *)route_list->l_head->le_data)->co_name[0] !=
              cn->cn_name[0]) ) 
            strcat(p->p_name,((Component *)route_list->l_head->le_data)->
                              co_name);
        else strcat(p->p_name,((Component *)route_list->l_tail->le_data)->
                               co_name);
 
         cn->cn_vpi=vpi;

         this_component = (Component *)route_list->l_head->le_data;
         other_component = (Component *)route_list->l_tail->le_data ;      

         n1= (Neighbor *)this_component->co_neighbors->l_head;
         n2= (Neighbor *)other_component->co_neighbors->l_head;

         if (n1->n_c->co_name[3] == n2->n_c->co_name[3])
            cn->cn_simple_connection=1;
         

           return ((caddr_t)cn); /* return something non-null */
#endif
}
 


















