
/*
   $Log: file.c,v $
   Revision 4.1  1996/09/24 19:36:09  tlin
   in world file, handles keywords node-graph and space-time

   Revision 4.0  1996/08/23 20:02:27  tlin
   bump to mickey mouse version

   Revision 3.25  1996/07/14 06:38:06  tlin
   typename -> _typename

   Revision 3.24  1996/07/14 06:12:06  tlin
   remove old read_world()



 */

/* file.c,v 1.0 1993/10/21 19:11:24 NIST */


/* Disk I/O routines, including read & write world file, and
   also write PS file to make hard copy of the world configuration */


#include "cisePort.h"
#include "sim.h"
#include "q.h"
#include "list.h"
#include "component.h"
#include "log.h"
#include "comptypes.h"
#include "event.h"
#include "eventdefs.h"
#include "route.h"
#include "main.h"

/* COMP_OBJECT pop_comp_window(); */



/* from simx.h */
list *route_table;


static void make_stuff(Component *c, Param *p, int x, int y, 
		       int width, int height)
{
/*     if (p->p_flags & MeterMask & M.doX) { */
/* 	if (pop_meter(c, p, x, y, width, height) == ERROR) */
/* 	    p->p_flags &= ~MeterMask;	/* Error--turn the meter flag off */
/*     } */

    /* FOR_DEMO: do not produce log files so that students cannot take it */
#   ifndef FOR_DEMO
    if (p->p_flags & LogMask)
	if (!log_param_create(c, p)) 
	    p->p_flags &= ~LogMask;	/* Error--turn logging flag off */
#   endif
}



/* Structure to store info about parameters for later use while creating the
   world. */
typedef struct _Param_info {
    struct _Param_info  *pi_next;	/* Pointer to put this thing in a queue. */
    short		pi_is_pflags;	/* TRUE if line in the file began "pflags" */
    char		*pi_value;	/* strcpy() of the value in the file. */
    int		        pi_flags;	/* value to be put in the p_flags field */
    int		        pi_type;	/* value for p_display_type field. */
    short		pi_x, pi_y;	/* X & Y locations for meters. */
    short		pi_height, pi_width;
} Param_info;




static int LineNum=0;
static Component *create_component_of_type(char *buf, char *key, list *clist)
{
    Component *c1=NULL;
    char cname[33], type[33];
    int  i, xpos=0, ypos=0;

    if (sscanf(buf, "%32s%32s%32s%d%d", key, cname, type, &xpos, &ypos)!=5) {
	printf("    Line %d: expects component <name> <type> <xpos> <ypos>\n", 
	       LineNum);
	return NULL;
    }
    for (i = 0; *component_types[i]._typename; i++)
	if (strcmp(component_types[i]._typename, type)==0)  {
	    c1 = (Component *)((*component_types[i].action)
			       (NULL,NULL,EV_CREATE,NULL,cname));
	    break;
	}
    if (*component_types[i]._typename==0) {
	printf("    Line %d: unknown type of component: %s\n", LineNum, type);
	return NULL;
    }

    le_addt(clist, c1);

    /* Make the component's window */
    if (M.doX) 
	c1->co_picture=(*component_types[i].icon)(c1, xpos, ypos);

    return c1;
}

int conclude_components(list *clist, queue *pqueue)
{
    /* Now, step through all the parameters of all the components in parallel
     * with all the saved Param_info structures.  Bash all the parameter's
     * flags and display_types.
     *
     * Note that the queue of queues of Param_info structures is being
     * dismembered and freed as each one is processed.
     */
    Component  *c1;
    queue      *piq;
    Param_info *pi;
    Param      *p;
    
    for (c1 = (Component *)clist->l_head, piq = (queue *)q_deq(pqueue);
	 c1 && piq;
	 c1 = (Component *)c1->co_next, piq = (queue *)q_deq(pqueue) ) {
      
	/* If it's a parameter with a value, initialize
	 * the param's value by calling p_input with the string & also 
	 * initialize its p_flags and p_display_type if they exist.
	 * Note that flags and/or display_type being zero in the world
	 * file is indistinguishable from their not being in the file at
	 * all (at this point in read_world()).
	 *
	 * If the world file is not a snapshot, there may be fewer parameters
	 * in the world file than in the component.  We deal with this by
	 * only setting parameters that have InputMask set in their flags.
	 */

	for (pi = (Param_info *)qe_deq(piq), p = (Param *)c1->co_params->q_head;
	     pi && p;
	     pi = (Param_info *)qe_deq(piq), p = p->p_next) {
	    if (pi->pi_is_pflags) {
		/*** It is "PFLAGS", set the p_flags & display_type */
		if (pi->pi_flags)  p->p_flags = pi->pi_flags;
		if (pi->pi_type)   p->p_display_type = pi->pi_type;
	
		/* Create any meters, logs, etc. */
		make_stuff(c1, p, pi->pi_x, pi->pi_y,
			   pi->pi_width, pi->pi_height);
	    } else {
		/* It's *not* FLAGS, so init the value & also check for p_flags.
		 * Skip any parameters on the list that do not want input.
		 */
		for (; p && ! (p->p_flags & InputMask); p = p->p_next);
		if (p) {
		    if (pi->pi_value) {
			(*p->p_input)(p, pi->pi_value, c1);
			free (pi->pi_value);
		    }
		    if (pi->pi_flags)  p->p_flags = pi->pi_flags;
		    if (pi->pi_type)   p->p_display_type = pi->pi_type;
		    
		    make_stuff (c1, p, pi->pi_x, pi->pi_y,
				pi->pi_width, pi->pi_height);
		}
	    }
	    free (pi);
	}

	/* Just in case... */
	for ( ; pi; pi = (Param_info *)qe_deq (piq)) {
	    if (pi->pi_value)	free (pi->pi_value);
	    free (pi);
	}
      
	/* Pop up the infowindow if necessary. */
	if (M.doX && c1->co_menu_up) {
	    /* Clear menu_up so as to not confuse X. */
	    c1->co_menu_up = FALSE;
	    /* pop_info_window(c1);  */
	}
	free (piq);
    }
}


static int get_param(char *buf, char *key, queue *piq)
{ 
    Param_info *pi = (Param_info *)sim_calloc(1, sizeof(Param_info));
    char       value[60];

    qe_addt (piq, (char *)pi);
	  
    pi->pi_is_pflags = (strcmp(key, "pflags")==0);
    if (pi->pi_is_pflags) { 
	sscanf(buf, "%s%x%x%hd%hd%hd%hd", key, 
	       &pi->pi_flags, &pi->pi_type, 
	       &pi->pi_x, &pi->pi_y, &pi->pi_width, &pi->pi_height);
    } else {
	sscanf(buf, "%s%60s%x%x%hd%hd%hd%hd", key, value, 
	       &pi->pi_flags, &pi->pi_type, 
	       &pi->pi_x, &pi->pi_y, &pi->pi_width, &pi->pi_height);
	pi->pi_value = sim_malloc(strlen (value) + 1);
	strcpy (pi->pi_value, value);
    }
}

static Component *find_comp(list *clist, char *name)
{
    Component *c1;
    for (c1 = (Component *)clist->l_head; c1!=NULL; c1=c1->co_next)
	if (strcmp(name, c1->co_name)==0)
	    return c1;
    printf("    Line %d: couldn't find component '%s' in list",LineNum, name);
    return NULL;
}

static int make_neighbor(char *buf, char *key, list *clist)
{
    char name1[32], name2[32];
    Component *c1, *c2;
    char      *result;

    if (sscanf(buf, "%s%32s%32s", key, name1, name2)!=3) {
	printf("    Line %d: %s needs two names.\n", LineNum, key);
	return -1;
    }
    c1 = find_comp(clist, name1);
    c2 = find_comp(clist, name2);
    if (c1==NULL || c2==NULL) 
	return -1;

    /* NEIGHBOR1 ==> only one call. */
    /* NEIGHBOR ==> two calls. */
    result = (*c1->co_action)(NULL, c1, EV_NEIGHBOR,NULL,c2);
    if (result && strcmp(key, "neighbor")==0) {
	(*c2->co_action)(NULL, c2, EV_NEIGHBOR,NULL,c1);
	/* I believe this is extra. -- tlin@cis	   
	   (*c1->co_action)(NULL, c1, EV_UNEIGHBOR,NULL,c2); */
    }
    if (M.doX) 
	show_neighbor_relation(c1,c2);
}


static int make_node_position(char *buf, list *clist)
{
    Component *c1;
    char key[32], name[32];
    int  x;
    
    sscanf(buf, "%s%s%d", key, name, &x);
   
    if ((strcmp(name,"phy1")!=0) && (strcmp(name,"phy2")!=0) && (strcmp(name,"phy3")!=0)) {
      printf("Make_node_position: it should be a physcial layer %s\n", name);
      exit(1);
    }
    
    c1 = find_comp(clist, name);
    
    if (c1 == NULL) {printf("Make_node_position: reutrn value is NULL"); exit(1);}
    c1->co_picture->bus_x = x;
    c1->co_picture->bus_y = 500;
}


static int make_route(char *buf, char *key, list *clist)
{
    int  oneway = FALSE;
    list *routel;
    Component *c1, *c2;
    char name[32];
    int  n;

    /* Build a list of the components */
    routel = l_create();
    /* Minimal error checking here--the components will log messages
       if there are errors */
    /* RG: CISE only enquene EV_MK_ROUTE to the phy layer of the 
       first item in the list */

    sscanf(buf, "%s%n", key, &n);   buf+=n;
    sscanf(buf, "%s%n", name, &n);  buf+=n;

    if ( (c1=find_comp(clist, name))==NULL )
	return -1;
    (*c1->co_action)(NULL, c1, EV_MAKE_ROUTE,NULL,buf);

    while (sscanf(buf, "%s%n", name, &n)==1) {
	buf+=n;
	if ( (c2=find_comp(clist, name))==NULL ) 
	    continue;
	if (find_neighbor(c1->co_neighbors, c2)==0) {
	    /* report unexpected error. need fix. -- tlin@cis */
	    printf("   Line %d: route to a non-neighbor %s\n",  
		   LineNum, name); 
	    continue;
	}
	l_addt(routel, (char *)c2);
	c1 = c2;
    }
	
    /* netsim show the Xcomponent here -- removed by tlin@cis */

    l_addt(route_table, routel);
}

static int process_line(char *buf, list *clist, queue *pqueue)
{
    char   key[32]="";
    static Component *c1=NULL;
    static queue     *piq=NULL;

    if (buf[0]=='#') return 0;         /* comment line     */
    if (sscanf(buf, "%31s", key)!=1)   /* check blank line */
	return 0;
    
    if        (strcmp(key,"component")==0) {
	c1=create_component_of_type(buf, key, clist);
	/* Create a new list of Param_info's for this component, and add
	   it to pqueue. */
	piq = q_create();
	q_addt(pqueue, (char *)piq);
    } else if (strcmp(key,"param")==0 || strcmp(key,"pflags")==0) {
	if (c1==NULL) {
	    printf("    Line %d: No component defined yet.\n", LineNum);
	    return -1;
	}
	get_param(buf, key, piq);
    } else if (strcmp(key,"neighbor")==0 ||strcmp(key,"neighbor1")==0) {
	make_neighbor(buf, key, clist);
    } else if (strcmp(key,"route")==0 || strcmp(key,"route1")==0) {
	make_route(buf, key, clist);
    } else if (strcmp(key,"node-graph")==0) {
	make_node_graph(buf);
    } else if (strcmp(key,"space-time")==0) {
	make_space_time(buf);
    } else if (strcmp(key,"node_position")==0) {
        make_node_position(buf,clist);
    }else {
	printf("    Line %d: unknown keyword %s in file. Skip.\n", LineNum, key);
	return -1;
    }
}



/* modified by tlin@cis
   no quote allowed, no continued line, no space allowed 
   in identifier, case sensitive now. In my view, these 
   features are 1. too fancy 2. not necessary 3. waste 
   of time to support.                                    */
list *read_world(char *name)
{
    FILE *fp;
    list *clist;
    queue *pqueue;		/* Parameter information list (of lists)*/
    char buf[1024];

    

    if (!(fp = fopen(name, "r")))  {
	printf("read_world:  couldn't open world file '%s'", name);
	free((char *)clist);
	return NULL;
    }

	clist = l_create();
    /* Plist is kept in parallel with clist.  It consists of a queue
       of Param_info structures for each component.  */
    pqueue = q_create();


    LineNum=1;
    while (fgets(buf, 1024, fp)!=NULL) {
	process_line(buf,clist,pqueue);
	LineNum++;
    }
    conclude_components(clist, pqueue);

    fclose(fp);
    if (M.doX) init_route_activity();
    return clist;
}


/****************************************************************
  Two routines to save the world in a file.
  One just saves components, routes, and neighbors, the other
  saves all of that along with the state of all meters, log files, 
  and infowindows.

  Both types of save can be read by read_world (above).
  Neither one is particularily fast or efficient, but that's okay.
*/

/* save_world & save_snapshot are just stubs that call save_whatever with
   the appropriate flag set. */
int save_world(list *clist, char *filename, list *route_table)
{
    /* clist: Linked list of all components */
    return(save_whatever(clist, filename, FALSE,route_table));
}

int save_snapshot(list *clist, char *filename, list *route_table)
{
    /* clist: Linked list of all components */
    return(save_whatever(clist, filename, TRUE,route_table));
}


int save_whatever(list *clist, char *filename, int snap, list *route_table)
{
    /* clist: Linked list of all components */
    /* snap: boolean flag */
    int x, y;
    Component *c1, *c2;
    Neighbor *n;
    Param *p;
    list *routel;
    l_elt *le;
    FILE *fp;
    l_elt *leo;
    l_elt *lei;
    list *route_list;

#   ifdef FOR_DEMO
    return TRUE;
#   endif
    
    /* Open file */
    if (!(fp = fopen(filename, "w")))  {
	dbg_write(DBG_ERR, NULL, 
		  "save_whatever: couldn't open %s for writing", filename);
	return(FALSE);
    }

    /* Write time & seed to the file. */
    if (snap) {
	fprintf (fp, "# Seed %d\n", M.seed);
	fprintf (fp, "# Time of snapshot (ticks) %u\n", ev_now() );
    }

/* $Revision: 4.1 $ */
/*
  fclose(fp);
  return(TRUE);
  */

  /*** First write out all components */
    for (c1 = (Component *)clist->l_head; c1; c1 = c1->co_next)  {
/*       component_xy(c1, &x, &y);	/* Position of the component */ 
/*       fprintf(fp, "component '%s' %s %d %d\n", c1->co_name, */
/* 	      component_types[c1->co_type]._typename, x, y); */
	/* State of infowindows if snapshot */
	if (snap && c1->co_menu_up)
	    fputs("infowindow\n", fp);
	/* Print out any parameters */
	for (p = (Param *)c1->co_params->q_head; p; p = p->p_next)  {
	    if (snap)  {		/* Save flags as well as value */
		if (p->p_flags & InputMask) {
		    fprintf(fp, "param %s %x %x",
			    (*p->p_make_short_text)(c1, p),
			    p->p_flags, p->p_display_type);
		    /*
		     * If there is a meter up, save the location & size.
		     */
		    if (p->p_flags & MeterMask) {
/* 		      METER *m = (METER *)p->p_my_picture; */
/* 		      fprintf (fp, " %d %d %d %d",  */
/* 			       m->x, m->y, m->width, m->height); */
		    }

		    /* Print the name as comment for human readers. */
		    fprintf (fp, "\t# %s\n", (*p->p_make_text)(c1, p));
		}
		else {
		    fprintf(fp, "pflags %x %x", p->p_flags, p->p_display_type);
		    /*
		     * If there is a meter up, save the location & size.
		     */
		    if (p->p_flags & MeterMask) {
/* 	      METER *m = (METER *)p->p_my_picture; */
/* 	      fprintf (fp, " %d %d %d %d", m->x, m->y, m->width, m->height); */
		    }
		    /* Print the name as comment for human readers. */
		    if (p->p_make_short_text) {
			fprintf(fp, "\t #%s: %s", p->p_name,
				(*p->p_make_short_text)(c1, p));
		    }
		    fprintf (fp, "\n");
		}
	    } else {
		if (p->p_flags & InputMask)
		    fprintf(fp, "param %s\t# %s\n",
			    (*p->p_make_short_text)(c1, p), (*p->p_make_text)(c1, p));
	    }
	}
	fputs("\n", fp);
    }

    /* Write out all neighborings (<- what a horrible word!) */
    for (c1 = (Component *)clist->l_head; c1; c1 = c1->co_next)
	for (n = (Neighbor *)c1->co_neighbors->l_head; n; n = n->n_next)  {
#if 0
	    if (snap && n->n_p != NULL)
	    {
		fprintf(fp, "neighbor1 '%s' '%s' %x %x", c1->co_name,
			n->n_c->co_name, n->n_p->p_flags, n->n_p->p_display_type);
		if (n->n_pp != NULL)
		{
		    fprintf(fp, " %x %x", n->n_pp->p_flags, n->n_pp->p_display_type);
		}
		fputs("\n", fp);
	    }
	    else
#endif   /* if 0 */
		fprintf(fp, "neighbor1 '%s' '%s'\n", c1->co_name, n->n_c->co_name);
	}

    fputs("\n", fp);

  /* And finally all routes. */
/*  for (c1 = (Component *)clist->l_head; c1; c1 = c1->co_next)
    if (c1->co_class == CONNECTION)  {
      s = (Socket *)(*c1->co_action)(NULL,c1,EV_GET_SOCKET,NULL,any_vpi,NULL);
      c2 = s->so_port;
      if (!c2) continue;*/		/* If no other port, no route anyway */
/*      if (routel = rt_find_route(c1))  {
	fprintf(fp, "route1");
*/
	/* Print out the route starting with the current end */
/*
	if (c1 == (Component *)routel->l_head->le_data)
	  for (le = routel->l_head; le; le = le->le_next)
	    fprintf(fp, " '%s'", ((Component *)le->le_data)->co_name);
	else
	  for (le = routel->l_tail; le; le = le->le_prev)
	    fprintf(fp, " '%s'", ((Component *)le->le_data)->co_name);
	fputs("\n", fp);
      }
      fflush(fp);
   }
*/

/* November 22,  93 */
/* Trying to save the contents of the routing_table */
  
    if (route_table)
    {
	for (leo=route_table->l_head;leo;leo=leo->le_next)
	{
	    fprintf(fp, "route1");
	    route_list = (list *)leo->le_data;
	    if (c1 == (Component *)route_list->l_head->le_data)
		for (lei=route_list->l_head;lei;lei=lei->le_next) 
		    fprintf(fp, " '%s'", ((Component *)lei->le_data)->co_name);
	    else
		for (lei = route_list->l_tail; lei; lei = lei->le_prev)
		    fprintf(fp, " '%s'", ((Component *)lei->le_data)->co_name);
	    fputs("\n", fp);
	}
	fflush(fp);
    }
   

    fclose(fp);
    return(TRUE);
}


/** routine to make a PostScript program to print out the world.
    It looks for the file "basicstuff.PS", copies it, and adds onto the
    end of it.  "basicstuff.PS" should include various support functions
    that I will write in PostScript.  */

create_PS_file(comps, name)
     list *comps;
     char *name;
{
    return(1);
}




