/* list.c,v 1.0 1993/10/21 19:11:46 NIST */
/* SHIV : cleaned up for CISE */
/*
 *
 * $Log: list.c,v $
 * Revision 4.0  1996/08/23 20:02:27  tlin
 * bump to mickey mouse version
 *
 * Revision 3.0  1996/05/30 14:40:54  tlin
 * bump to 3.0; before Tk adition
 *
 * Revision 2.1  1996/05/28 05:57:31  tlin
 * bump to v 2.1; before tcl/tk conversion
 *
 * Revision 0.1  1996/05/27 22:51:28  tlin
 * test
 *
 * Revision 0.0  96/04/26  15:34:24  goyal
 * Lab1
 * 
 * Revision 1.0  96/04/24  15:01:44  shivkuma
 * Lab 1
 * 
 * Revision 1.1  96/04/21  23:59:59  shivkuma
 * Version Before Putting in User-Friendly names
 * 
 * Revision 0.0  96/02/21  00:44:00  goyal
 * Initial Revision
 * 
 *
 *
 */


/* list.c:  Rest of list routines */
/*LINTLIBRARY*/

/* General-purpose list manipulation routines.  Contains the following
 * routines:
 *	l_add{t,h}	Add element to head or tail if list.
 *	l_adda		Add element after another in the list.
 *	l_del		Delete an element from a list.
 *	l{e}_remh	remove and return head of list.
 *	l{e}_remt	remove and return tail of list.
 *	l_find		determine whether an element is in a list.
 *	l_create	create a list.
 *	lq_delete	Free all storage in either a list or a queue.
 *	l_obliterate	Free all storage (including user data) in a list.
 * All other general-purpose list manipulation routines are macros
 * defined in "list.h".
 */

#include <sys/types.h>
#include <stdio.h>
#include "list.h"

extern char *sim_calloc();

#define	TRUE	1
#define	FALSE	0

/* The following macros implement most of the common list operations */

/* Add an element to the head of the list */

void le_addh(list *l, l_elt *elt)  { 
	if ((l)->l_head == 0) 
	  (l)->l_tail = (l_elt *)(elt); 
	else 
	  (l)->l_head->le_prev = (l_elt *)(elt); 
	((l_elt *)(elt))->le_next = (l)->l_head; 
	((l_elt *)(elt))->le_prev = 0; 
	(l)->l_head = (l_elt *)(elt); 
	if (++((l)->l_len) > (l)->l_max) 
	  (l)->l_max = (l)->l_len; 
}
			   
/* Add an element to the tail of a list */

void le_addt(list *l, l_elt *elt)  { 
	if ((l)->l_tail == 0) 
	  (l)->l_head = (l_elt *)(elt);
	else 
	  (l)->l_tail->le_next = (l_elt *)(elt); 
	((l_elt *)(elt))->le_prev = (l)->l_tail; 
	((l_elt *)(elt))->le_next = 0; 
	(l)->l_tail = (l_elt *)(elt); 
	if (++((l)->l_len) > (l)->l_max) 
	  (l)->l_max = (l)->l_len; 
}

/* Add an element after a specified element in the list.  If prev == */
/* &l->l_head, can be used to add an element to the head of the queue */

void le_adda(list *l, l_elt *prev, l_elt *new1)  { 
	if ((l)->l_head == 0)  {		            /*** Empty */
	  ((l_elt *)(new1))->le_prev = ((l_elt *)(new1))->le_next = 0; 
	  (l)->l_tail = (l_elt *)(new1); 
	} else if ((l)->l_tail == (l_elt *)(prev))  {         /*** At end */ 
	  ((l_elt *)(new1))->le_prev = (l_elt *)(prev); 
	  (l)->l_tail = (l_elt *)(new1); 
	} else if ((l_elt *)(&l->l_head) == (l_elt *)(prev))  { /*** At front */ 
	  ((l_elt *)(new1))->le_prev = 0; 
	  (l)->l_head->le_prev = (l_elt *)(new1); 
	} else  {					    /*** Middle */ 
	  ((l_elt *)(new1))->le_prev = (l_elt *)(prev); 
	  ((l_elt *)(prev))->le_next->le_prev = (l_elt *)(new1); 
	} 
	((l_elt *)(new1))->le_next = ((l_elt *)(prev))->le_next; 
	((l_elt *)(prev))->le_next = (l_elt *)(new1);  
	if (++((l)->l_len) > (l)->l_max)  
	  (l)->l_max = (l)->l_len;  
}


/* Delete a given element. */

void le_del(list *l, l_elt *elt)  { 
	if (((l_elt *)(elt))->le_next)  
	  ((l_elt *)(elt))->le_next->le_prev = ((l_elt *)(elt))->le_prev;  
	else  
	  (l)->l_tail = ((l_elt *)(elt))->le_prev;  
	if (((l_elt *)(elt))->le_prev)  
	  ((l_elt *)(elt))->le_prev->le_next = ((l_elt *)(elt))->le_next;  
	else  
	  (l)->l_head = ((l_elt *)(elt))->le_next;  
	if (--(l)->l_len < (l)->l_min)  
	  (l)->l_min = (l)->l_len;  
}


l_elt *l_addh(l, elt)
     register list *l;
     register char *elt;
{
  register l_elt *new = (l_elt *)sim_calloc(1, sizeof(l_elt));

  new->le_data = elt;
  le_addh(l, new);
  return(new);
}

l_elt *
l_addt(l, elt)
     register list *l;
     register char * elt;
{
  register l_elt *new = (l_elt *)sim_calloc(1, sizeof(l_elt));

  new->le_data = elt;
  le_addt(l, new);
  return(new);
}


/* Procedure to determine if an element is in a list.
   Returns NULL if not found, and the pointer to the *l_elt* if found. */
l_elt *
l_find(l, elt)
     register list *l;
     register char * elt;
{
  register l_elt *tmp;

  for (tmp = l->l_head; !tmp || tmp->le_data != elt; tmp = tmp->le_next)
    if (!tmp)
      return(NULL);
  return(tmp);
}


l_elt *
l_adda(l, prev, new1)
     register list *l;
     register char *prev, *new1;
{
  register l_elt *newelt;
  register l_elt *found;
  

  found = l_find(l, prev);
  if (found)  {
    newelt = (l_elt *)sim_calloc(1, sizeof (l_elt));
    newelt->le_data = new1;
    le_adda(l, found, newelt);
  }
  return(newelt);
}

l_del(l, elt)
     register list *l;
     register char *elt;
{
  register l_elt *found;

  found = l_find(l, elt);
  if (found)  {
    le_del(l, found);
    free((char *)found);
    return(TRUE);
  }
  return(FALSE);
}

char *
l_remh(l)
     register list *l;
{
  register l_elt *head = le_remh(l);
  register char * data = NULL;

  if (head)  {
    data = head->le_data;
    free((char *)head);
  }
  return(data);
}

char *
l_remt(l)
     register list *l;
{
  register l_elt *tail = le_remh(l);
  register char *data = NULL;

  if (tail)  {
    data = tail->le_data;
    free((char *)tail);
  }
  return(data);
}

l_elt *
le_remh(l)
     register list *l;
{
  register l_elt *le;

  if (l->l_head)  {
    le = l->l_head;
    l->l_head = le->le_next;
    if (le->le_next)
      le->le_next->le_prev = NULL;
    else
      l->l_tail = NULL;
    if (--l->l_len < l->l_min)
      l->l_min = l->l_len;
    return(le);
  } else
    return(NULL);
}


l_elt *
le_remt(l)
     register list *l;
{
  register l_elt *le;

  if (l->l_tail)  {
    le = l->l_tail;
    l->l_tail = le->le_prev;
    if (le->le_prev)
      le->le_prev->le_next = NULL;
    else
      l->l_head = NULL;
    if (--l->l_len < l->l_min)
      l->l_min = l->l_len;
    return(le);
  } else
    return(NULL);
}


list *
l_create()
{
  return((list *)sim_calloc(1, sizeof(list)));
}



/* Free a list or queue entirely.  This **DOES NOT** free any of the
   storage pointed to by the list or queue elements. */

void
lq_delete(lq)
     register list *lq;
{
  register l_elt *le, *next;

  for (le = lq->l_head; le; le = next)  {
    next = le->le_next;
    free((char *)le);
  }
  free((char *)lq);
}

/* Remove all elements from a list or queue.  Similar to lq_delete, but
   doesn't free the list itself.  */
void
lq_clear(lq)
     register list *lq;
{
  register l_elt *le, *next;

  for (le = lq->l_head; le; le = next)  {
    next = le->le_next;
    free((char *)le);
  }
  lq->l_len = 0;
  lq->l_max = lq->l_min = 0;
  lq->l_head = (l_elt *)NULL;
  lq->l_tail = (l_elt *)NULL;
}


/* Free everything associated with a list.
   *** NOTE *** This is __NOT__ the opposite of l_create.  l_create only
   allocates memory for the list header, not for all the elements in the
   list.  This function frees the elements *and* the data pointed to by the
   elements as well as the header, and is
   provided as a service to the user.  Use at your own risk. */

void
l_obliterate(l)
     register list *l;
{
  register l_elt *le, *next;

  for (le = l->l_head; le; le = next)  {
    next = le->le_next;
    free((char *)le->le_data);
    free((char *)le);
  }
  free((char *)l);
}

    
