#include "cisePort.h"
extern "C" {
#include "q.h"
#include "list.h"
#include "hash.h"
#include "sim.h"
#include "eventdefs.h"
#include "component.h"
#include "main.h"
#include "sim_tk.h"
#include "route_activity.h"
#include "string.h"
};




#define SEEXNOW XFlush(M.display) 

/* ------------------------------------------------------------------

   route_activity() by tlin@cis 3/96
   draw blocks for packets instead of flashing the route only

   ------------------------------------------------------------------ */


#define KEYSIZE  20
#define MAXDASH  20  
#define MAXROUTE 14   /* see also xinit.c */


#define MAX_ACTIVE_ROUTES 97   /* for hash table; prime # preferred */
#define MAX_PACKET_IN_LINK 50  /* MAX # of packets in one direction of link */
#define PACKET_LENGTH 50       /* Packget Length */

/* Reference to sim.tcl: proc draw_bus {} */
#define BUS_HEAD_X  100
#define BUS_HEAD_Y  500
#define BUS_END_X   600
#define BUS_END_Y   500
#define BUS_DENSITY 50         /* 1(km) = 50 (pixel) */
#define BlockW      5          /* Each block is 5 pixel wide*/
#define BlockH      5
#define Next_draw_block_time 500 /* packet will take 5 ns to go through 1km */ 
                                 /* 50/5 * 500 (ms) = 5000 (ms) = 5(ns) */     
typedef struct { 
    int   from_x, from_y, to_x, to_y, last_x, last_y;
    int   length;		/* Packet Length: used in Ethernet */
    int   age, error;
    ulong delay;
    ulong start_time;
    char   color[20];           /* Ethernet: green: normal, red: collision */
} packet_activity_t;

typedef struct { 
    packet_activity_t packet[MAX_PACKET_IN_LINK];
    int    x1, y1, x2, y2;        /* from (x1,y1) to (x2,y2)       */
    int    h, t, qn;              /* head, tail, #element of queue */
} route_activity_t;

static Htable HTRouteActivity=NULL;  /* hash the address of RouteActivity */


/* event_array for undraw packet event */
/* keep the event point of undraw packet event */
#define  MaxEvent 20
Event *event_array[MaxEvent];
int      CurrentNumEvent = 0;

int init_route_activity()
{
    delete_route_activity();
    HTRouteActivity=
	ht_create(MAX_ACTIVE_ROUTES, KEYSIZE, 
		  sizeof(route_activity_t));      
	return 0;   
}

int delete_route_activity()
{
    if (HTRouteActivity!=NULL) 
	ht_obliterate(HTRouteActivity);
    HTRouteActivity=NULL;

	return 0;
}

Pixmap pix_bomb=0;


void make_new_route(XComponent *from, XComponent *to,  
		    route_activity_t *rt);

void make_new_bus_route(XComponent *from, XComponent *to, 
		    route_activity_t *rt, char *src_name);
static void conn_point(int x, int y, int w, int h, int which, int *px, int *py);


int check_collision(packet_activity_t *pkt )
{
  /* Purpose: checking if the packet should be changed color because of */
  /*          the collision.                                            */
  /*                                                                    */
  /* Algorithm:                                                         */
  /*          If the packet color == red, that means there has been     */
  /*                   collsions happened on the packet.                */ 
  /*                     So we need not check it                        */
  /*          However, if it is not true. We need to check collsions    */
  /*                    We check the collision_array[], if there is     */
  /*                    any collsion between the current packet position*/
  /*                    and the starting packet position. We set the    */
  /*                    color to red.                                   */
  
  if (strcmp(pkt->color, "red") == 0)
    return 0;

  // check if there is a collision happening between packet start position and
  // current position.
  for (int i=0; collision_array[i]!=-1;i++) { 
      
     if (((pkt->last_x - pkt->from_x)/((collision_array[i]-pkt->from_x)*1.0) >= 0.0) &&
 	(abs(pkt->last_x - pkt->from_x)>=abs(collision_array[i]-pkt->from_x))) {
       strcpy(pkt->color,"red");
       return 0;
     }

  }
  
  return 0;
}

int check_interface_array()
{
  /* 1. check if there is a new collision in the interface_array[], */
  /* 2. trasfer the collision from (km) to the (xpos) on the screen */
  /* 3. Store it in the collision_array[]                           */

  int index =0;

  for (index =0; collision_array[index]!= -1; index++); // find the empty space

  for (int i=0; interface_array[i]!=-1.0; i++) {
    // transfer from km to pixel on screen
    collision_array[index]=(int)(interface_array[i]*BUS_DENSITY+BUS_HEAD_X); 
    index++;
    interface_array[i]=-1.0; // reset
  }

  return 0;
}

int extract_depth(char *name)
{				
  //  Convert char N to the corresponding decimal N
  //  And the format of name must be "phyN", N is a number
  //  It should be modified later. 
  
  return name[3]-'0';
}




int draw_packet_from_bus_to_dest(Component *src, Component *dest, int delete_packet_flag)
{

  XComponent *from=(XComponent *) src->co_picture;
  XComponent *to  =(XComponent *)dest->co_picture;

  int depth, p_to;
  int x1, x2, y1, y2;
  int value, last_value = 1000000;	// big number
    
  char              key[KEYSIZE]="";
  route_activity_t  *rt;
  packet_activity_t *pkt;
  int               j, h, qn;
  int               x, y, fx, fy, tx, ty;
  char*             color;
  float dt;

  /* get the packet color */
  from->bus_dest_x = BUS_HEAD_X;
  from->bus_dest_y = BUS_HEAD_Y;
  
  strcpy(key,                     src->co_name);
  strcat(key,                     dest->co_name);
  if ((to->bus_x - from->bus_x)>=0)
    strcat(key,                     "1");
  else
    strcat(key,                     "0");

  if (delete_packet_flag == 0)
      strcat(key,                     "0");
    else
      strcat(key,                     "1");
 

  rt=(route_activity_t*)ht_find(HTRouteActivity, key);
  if (rt==NULL) {
    printf("Internal error: update non-existing route.\n");
    return -1;
  }

  
  qn=rt->qn;
  for (j=0, h=rt->t; j<qn; j++, h=(h+1)%MAX_PACKET_IN_LINK) {
    pkt=&(rt->packet[h]);   
    // find the correct color of the nearest packet around the destination
    value = abs(pkt->last_x - to->bus_x);
    if (value < last_value) {
      last_value = value;
      color = pkt->color;
    }

	
  }

 
	
  // how deep under the bus this packet should be shown on the screen?    
  depth = extract_depth(src->co_name);
  switch (depth) {
  case 1:
    p_to=16; break;
  case 2:
    p_to=17; break;
  case 3:
    p_to=18; break;
  case 4:
    p_to=19; break;
  default:
    printf("Error: draw_packet_from_bus_to_dest()\n");
    printf("Error: Physcial layer name should be PhyN, 1<=N<=4\n");
    exit;
  }
  
  conn_point(  to->x,  to->y,  to->w,  to->h,     p_to, &x1, &y1);
  conn_point(to->bus_x,to->bus_y,from->w,from->h, p_to, &x2, &y2);

    
  if (!delete_packet_flag) {
    draw_packet(src->co_name, color, x1, to->y,BlockW, y2-dest->co_picture->y); 
  }
  else 
      delete_packet(src->co_name, x1, to->y);
         

  return 0;
		   
}

int start_sending_packet(Component *src, Component *dest)
{
  
  /*
   * 1. draw_block from src->src_attach
   * 2. route_activity from src_attach to bus_end
   * 3. route_activity from src_attach to bus_head
   * 4. draw_block from dest_attach -> dest at the appropriate time
   */
   int trans_time;
   int delete_packet_flag=0;
   
  /*
   * Calculate the needed time to transfer from src to dest on bus,
   * then issue the event to draw block from bus to dest
   */
   XComponent *from=(XComponent *) src->co_picture;
   XComponent *to  =(XComponent *)dest->co_picture;

   int x1, x2, y1, y2;

  
   conn_point(from->bus_x,  from->bus_y,  from->w,  from->h, 7, &x1, &y1);
   conn_point(to->bus_x,    to->bus_y,    to->w,    to->h,  12, &x2, &y2);

   // calculate transmit time
  trans_time = (x2 - x1)/BlockW*Next_draw_block_time;
  if (trans_time < 0) trans_time = -trans_time;
 
  bus_activity(src,dest,"green" ,0,delete_packet_flag);

  
  event_array[CurrentNumEvent]=ev_enqueue(EV_DRAW_PACKET_FROM_BUS_TO_DEST_ACTIVITY, src, dest, \
					  ev_now()+trans_time+1, src->co_action, NULL, \
					  (char*)delete_packet_flag);
  CurrentNumEvent++;		// increase the events recorder
  
   

  return 0;
}

int stop_sending_packet(Component *src, Component *dest)
{

  
  /*
   * 1. draw_block from src->src_attach
   * 2. route_activity from src_attach to bus_end
   * 3. route_activity from src_attach to bus_head
   * 4. draw_block from dest_attach -> dest at the appropriate time
   */
   int trans_time;
   int delete_packet_flag = 1;
   
  /*
   * Calculate the needed time to transfer from src to dest on bus,
   * then issue the event to draw block from bus to dest
   */
   XComponent *from=(XComponent *) src->co_picture;
   XComponent *to  =(XComponent *)dest->co_picture;

   int x1, x2, y1, y2;

  
   conn_point(from->bus_x,  from->bus_y,  from->w,  from->h, 7, &x1, &y1);
   conn_point(to->bus_x,    to->bus_y,    to->w,    to->h,  12, &x2, &y2);

   // calculate transmit time
  trans_time = (x2 - x1)/BlockW * Next_draw_block_time;
  if (trans_time < 0) trans_time = -trans_time;
 
  bus_activity(src,dest,"green",0, delete_packet_flag);
  ev_enqueue(EV_DRAW_PACKET_FROM_BUS_TO_DEST_ACTIVITY, src, dest, \
	     ev_now()+trans_time,  src->co_action, NULL, (char *)delete_packet_flag);


  return 0;
}


int bus_activity(Component *src, Component *dest, 
		       char* color, int error, int delete_packet_flag)
{
    
    XComponent *from=(XComponent *) src->co_picture;
    XComponent *to  =(XComponent *)dest->co_picture;

    char       key[KEYSIZE]="";
    route_activity_t  *rt0;
    route_activity_t  *rt1; 
    packet_activity_t *pkt0; 
    packet_activity_t *pkt1;
    int           length;

    
    /*
     * BUS_HEAD  ------------------------------- BUS_END
     *            <--------rt0  rt1 ---------->
     */
    from->bus_dest_x = BUS_HEAD_X;
    from->bus_dest_y = BUS_HEAD_Y;
    strcpy(key,                     src->co_name);
    strcat(key,                     dest->co_name);
    strcat(key,                     "0");
    if (delete_packet_flag == 0)
      strcat(key,                     "0");
    else
      strcat(key,                     "1");
    

    rt0=(route_activity_t*)ht_find(HTRouteActivity, key);

    if (rt0==NULL) {
	route_activity_t new_rt;
	make_new_bus_route(from, to, &new_rt, src->co_name);
	ht_insert(HTRouteActivity, key, (char*)&new_rt);	
	rt0=(route_activity_t*)ht_find(HTRouteActivity, key);
    }

    
    from->bus_dest_x = BUS_END_X;
    from->bus_dest_y = BUS_END_Y;
    
    strcpy(key,                     src->co_name);
    strcat(key,                     dest->co_name);
    strcat(key,                     "1");
    if (delete_packet_flag == 0)
      strcat(key,                     "0");
    else
      strcat(key,                     "1");

    rt1=(route_activity_t*)ht_find(HTRouteActivity, key);

    if (rt1==NULL) {
      route_activity_t new_rt;
      make_new_bus_route(from, to, &new_rt, src->co_name);
      ht_insert(HTRouteActivity, key, (char*)&new_rt);	
      rt1=(route_activity_t*)ht_find(HTRouteActivity, key);
    }

    
    if (rt0->qn>=MAX_PACKET_IN_LINK) {
       	return -1;
    }
    
    if (rt1->qn>=MAX_PACKET_IN_LINK) {
       	return -1;
    }
    
    
    pkt0=&(rt0->packet[rt0->h++]);
    rt0->h = (rt0->h) % MAX_PACKET_IN_LINK;
    rt0->qn++; 
    pkt0->from_x=rt0->x1;
    pkt0->from_y=rt0->y1;
    pkt0->to_x  =rt0->x2;
    pkt0->to_y  =rt0->y2;
    pkt0->delay =1;
    strcpy(pkt0->color,color);
    pkt0->start_time=ev_now();
    pkt0->last_x=pkt0->from_x;
    pkt0->last_y=pkt0->from_y;
    pkt0->age=0;
    pkt0->error=-1;
    if (error) pkt0->error=rand()%4+1;

    
    pkt1=&(rt1->packet[rt1->h++]);
    rt1->h = (rt1->h) % MAX_PACKET_IN_LINK;
    rt1->qn++; 
    pkt1->from_x=rt1->x1;
    pkt1->from_y=rt1->y1;
    pkt1->to_x  =rt1->x2;
    pkt1->to_y  =rt1->y2;
    pkt1->delay =1;
    strcpy(pkt1->color ,color);
    pkt1->start_time=pkt0->start_time;
    pkt1->last_x=pkt1->from_x;
    pkt1->last_y=pkt1->from_y;
    pkt1->age=0;
    pkt1->error=-1;
    if (error) pkt1->error=rand()%4+1;
   

    /* 
     * Draw long block between phy layer to bus, 
     *  packet transfer from phy layer to bus.
     */

    
    if (!delete_packet_flag) {
      draw_packet(src->co_name, color, pkt0->from_x, src->co_picture->y, \
		  BlockW, pkt0->from_y-src->co_picture->y); 
      draw_packet(src->co_name, color, pkt0->from_x, pkt0->from_y, BlockW, BlockH);
    }
    else {
      delete_packet(src->co_name, pkt0->from_x, src->co_picture->y); 
      delete_packet(src->co_name, pkt0->from_x, pkt0->from_y);
    }
    ev_enqueue(EV_UPDATE_BUS_ACTIVITY, src, dest, ev_now(),
	       src->co_action, NULL, (char *)delete_packet_flag);
    

    return 0;
}


int update_bus_routes(Component *src, Component *dest, int delete_packet_flag)
{
    XComponent        *from=(XComponent *) src->co_picture;
    XComponent        *to  =(XComponent *)dest->co_picture;
    char              key[KEYSIZE]="";
    route_activity_t  *rt0, *rt1;
    packet_activity_t *pkt0, *pkt1;
    int               j, h, qn;
    int               x, y, fx, fy, tx, ty;
    char*             color;
    float dt;

    from->bus_dest_x = BUS_HEAD_X;
    from->bus_dest_y = BUS_HEAD_Y;
    
    strcpy(key,                     src->co_name);
    strcat(key,                     dest->co_name);
    strcat(key,                     "0");
    if (delete_packet_flag == 0)
      strcat(key,                     "0");
    else
      strcat(key,                     "1");

    rt0=(route_activity_t*)ht_find(HTRouteActivity, key);
    if (rt0==NULL) {
	printf("Internal error: update non-existing route.\n");
	return -1;
    }

        
    from->bus_dest_x = BUS_END_X;
    from->bus_dest_y = BUS_END_Y;

    strcpy(key,                     src->co_name);
    strcat(key,                     dest->co_name);
    strcat(key,                     "1");
    if (delete_packet_flag == 0)
      strcat(key,                     "0");
    else
      strcat(key,                     "1");

    rt1=(route_activity_t*)ht_find(HTRouteActivity, key);
    if (rt1==NULL) {
        printf("Internal error: update non-existing route.\n");
	return -1;
    }

    

    qn=rt0->qn;
    for (j=0, h=rt0->t; j<qn; j++, h=(h+1)%MAX_PACKET_IN_LINK) {
	
	pkt0=&(rt0->packet[h]);

	/* destroy the packet */
	if (pkt0->last_x < BUS_HEAD_X) {	  
	  rt0->qn--;
	  rt0->t=(rt0->t+1)%MAX_PACKET_IN_LINK;
	  continue;
        }

		
        /* check if there is a collsion happening */
	pkt0->last_x-=BlockW*2;
	
	check_interface_array();
	check_collision(pkt0);

	color = pkt0->color;
	pkt0->last_x+=BlockW;
	
	if (!delete_packet_flag)
	  draw_packet(src->co_name, color, pkt0->last_x, pkt0->last_y, BlockW, BlockH);
	else
	  delete_packet(src->co_name, pkt0->last_x, pkt0->last_y);



    }
    
    
    qn=rt1->qn;
    for (j=0, h=rt1->t; j<qn; j++, h=(h+1)%MAX_PACKET_IN_LINK) {

	
	pkt1=&(rt1->packet[h]);

	if (pkt1->last_x > BUS_END_X) {
          /* destroy packet */	
	  rt1->qn--;
	  rt1->t=(rt1->t+1)%MAX_PACKET_IN_LINK;
	  continue;
        }

			
	/* check if there is a collsion happening */
	
	check_interface_array();
	check_collision(pkt1);
	
	pkt1->last_x+=BlockW;
	color = pkt1->color;
	
	
	if (!delete_packet_flag)
	  draw_packet(src->co_name, color, pkt1->last_x, pkt1->last_y, BlockW, BlockH);
	else
	  delete_packet(src->co_name, pkt1->last_x, pkt1->last_y);



    }
  

    if (rt0->qn!=0 || rt1->qn!=0) {   /* some packets still needs to be taken care of */
	tick_t next_ev_time=ev_now()+Next_draw_block_time;
	if (next_ev_time<ev_now()) /* make the "stub" of block disappear asap */
	    next_ev_time=ev_now()+1;
	ev_enqueue(EV_UPDATE_BUS_ACTIVITY, src, dest, next_ev_time,
		   src->co_action, NULL, (char *)delete_packet_flag);
    }

	return 0;
}


int route_activity(Component *src, Component *dest, tick_t delay, 
		       int color, int error)
{

    return 0;
}

int update_routes(Component *src, Component *dest)
{

	return 0;
}



static void conn_point(int x, int y, int w, int h, int which, int *px, int *py)
{
    *px=x; *py=y;

    switch (which) {
    case 0:
	*px+=w/2; break;
    case 1:
	*px+=w; *py+=h/2; break;
    case 2:
	*px+=w/2; *py+=h; break;
    case 3:
	*py+=h/2; break;
    case 4:
	*px+=w/4; break;
    case 5:
	*px+=w*3/4; break;
    case 6:
	*px+=w; *py+=w/4; break;
    case 7:
	*px+=w; *py+=w*3/4; break;
    case 8:
	*px+=w*3/4; *py+=w; break;
    case 9:
	*px+=w/4; *py+=w; break;
    case 10:
	*py+=w*3/4; break;
    case 11:
	*py+=w/4; break;
    case 12:
        *px+=w/4; *py+=h/4; break;
    case 13:
        *px+=w/4; *py+=h/2; break;
    case 14:
        *px+=w/4; *py+=h*3/4; break;
    case 15:
        *px+=w/4; *py+=h; break;
    case 16:
        *px-=w/4; *py+=h/4; break;
    case 17:
        *px-=w/4; *py+=h/2; break;
    case 18:
        *px-=w/4; *py+=h*3/4; break;
    case 19:
        *px-=w/4; *py+=h; break;
    }
}


void make_new_bus_route(XComponent *from, XComponent *to, 
		    route_activity_t *rt, char *src_name)
{
    int dx, dy, p_from, p_to;
    int depth;

    rt->h=rt->t=rt->qn=0;
  
    dx=to->bus_x-from->bus_x;
    dy=to->bus_y-from->bus_y;


    if (dy!=0) {
      printf("Error:make_new_bus_route(), dy!=0\n");
      exit ;
    }

    // how deep under the bus this packet should be shown on the screen?    
    depth = extract_depth(src_name);
    switch (depth) {
    case 1:
      p_from=12; p_to=16; break;
    case 2:
      p_from=13; p_to=17; break;
    case 3:
      p_from=14; p_to=18; break;
    case 4:
      p_from=15; p_to=19; break;
    default:
      printf("Error: make_new_bus_route()\n");
      printf("Error: Physcial layer name should be PhyN, 1<=N<=4\n");
      exit;
    }
    
    

    /* set up from (rt->x1, rt->y1) to (rt->x2, rt->y2), which is end of bus */
    conn_point(from->bus_x,from->bus_y,from->w,from->h,p_from,&rt->x1,&rt->y1);
    conn_point(from->bus_dest_x,from->bus_dest_y, from->w, from->h,p_to,  &rt->x2,&rt->y2);

}

//
// this refresh_route_activity() is ugly looking. Can someone
// make it better? -- tlin@cis
//
#include <tk.h>
struct {
    int               x1, y1, x2, y2;
} Exp;  /* for exposure event */


void refresh_route_activity1(ClientData clientData);
void refresh_route_activity2();

void refresh_route_activity(ClientData clientData, XEvent *eventPtr)
{
    XExposeEvent      *xevent=&eventPtr->xexpose;
    
    Exp.x1=xevent->x; Exp.x2=Exp.x1 + xevent->width;
    Exp.y1=xevent->y; Exp.y2=Exp.y1 + xevent->height;

    if (xevent->count!=0) return;
    Tk_DoWhenIdle(refresh_route_activity1, 0);
}

void refresh_route_activity1(ClientData clientData)
{
    static c=0;
    
    if (c%3)    // experimented value
	Tk_DoWhenIdle(refresh_route_activity1, 0);
    else
	refresh_route_activity2();
}

void refresh_route_activity2()
{
 
}

int flash_component(Component *c)
{
    return 0;
}


int reset_collision_arrays()
{
	int i;
    for (i=0; i<MaxCollision; i++){
	collision_array[i]=-1;
	interface_array[i]=-1.0;
    }

    // delete all packets on screen
    delete_all_packet();
    
    // dequeue the undraw_packet_events from event_queue.
    for (i=0; i<CurrentNumEvent; i++) {
	ev_dequeue(event_array[i]);
	event_array[i]=NULL;
    }
    CurrentNumEvent = 0;	// global variable

	return 0;
}





















