/* SHIV : cleaned up for CISE */

/* subr.c,v 1.0 1993/10/21 19:08:01 NIST */
/*
 *
 * $Log:	subr.c,v $
 * Revision 4.0  96/08/23  20:02:27  tlin
 * bump to mickey mouse version
 * 
 * Revision 3.1  1996/05/30 21:50:40  tlin
 * add usleep() for non-BSD
 *
 * Revision 3.0  1996/05/30 14:40:54  tlin
 * bump to 3.0; before Tk adition
 *
 * Revision 2.8  1996/05/30 11:40:19  tlin
 * find_neighbor()
 *
 * Revision 2.7  1996/05/30 10:44:28  tlin
 * more prototping, make_text() ...
 *
 * Revision 2.6  1996/05/30 10:26:30  tlin
 * more prototyping, stdarg.h
 *
 * Revision 2.5  1996/05/29 23:16:34  tlin
 * prototyping
 *
 * Revision 2.4  1996/05/29 16:20:20  tlin
 * dgb_write() redo
 *
 * Revision 2.3  1996/05/28 23:20:39  tlin
 * fixed ifdef -> if defined(.)
 *
 * Revision 2.2  1996/05/28 23:01:54  tlin
 * change #ifdef HP700 to hppa
 *
 * Revision 2.1  1996/05/28 05:57:31  tlin
 * bump to v 2.1; before tcl/tk conversion
 *
 * Revision 0.1  1996/05/27 22:51:46  tlin
 * test
 *
 * Revision 0.0  96/04/26  15:34:33  goyal
 * Lab1
 * 
 * Revision 1.0  96/04/24  15:02:07  shivkuma
 * Lab 1
 * 
 * Revision 1.1  96/04/22  00:00:10  shivkuma
 * Version Before Putting in User-Friendly names
 * 
 * Revision 0.0  96/02/21  00:44:04  goyal
 * Initial Revision
 * 
 *
 *
 */


/* File for the simulator that contains random functions that don't belong
   anywhere else.  */

#include "cisePort.h"
#include "sim.h"
#include "q.h"
#include "list.h"
#include "component.h"
#include "log.h"
#include "eventdefs.h"
#include "pdu.h"
#include "comptypes.h"
#include "subr.h"

/*********************************************
  Routines for various types of parameter I/O.
*/

/* Simplest text routine */
char *make_text(Component *c, Param *p)
{
    return(p->p_name);
}

char *make_name_text(Component *c, Param *p)
{
    return(c->co_name);
}

char *make_short_name_text(Component *c, Param *p)
{
    static char buf[50];
    
    sprintf(buf, "'%s'", c->co_name);
    return(buf);
}


/* Input routine for the parameter that is the component's name. */
int param_input_name(Param *p, char *s, Component *c)
{
    strcpy(c->co_name, s);
    return(TRUE);
}


/* Return a string that consists of the name of the passed parameter,
   followed by a colon and its value.  The pointer returned is to a static 
   data area, and so the contents will change with each call.  */
char *make_int_text(Component *c, Param *p)
{
    static char b[60];
    
    sprintf(b, "%s: %d", p->p_name, p->u.i);
    return(b);
}


char *make_short_int_text(Component *c, Param *p)
{
  static char b[40];

  sprintf(b, "%d", p->u.i);
  return(b);
}


double int_calc(Component *c, Param *p)
{
    return (double)p->u.i;
}


/* Same routines for doubles */
char *make_double_text(Component *c, Param *p)
{
  static char b[100];

  sprintf(b, "%s: %g", p->p_name, p->u.d);
  return(b);
}


/*ARGSUSED*/
char *make_short_double_text(Component *c, Param *p)
{
  static char b[40];

  sprintf(b, "%g", p->u.d);
  return(b);
}

/*ARGSUSED*/
double double_calc(Component *c, Param *p)
{
  return (double)p->u.d;
}


/* Return a string that consists of the name of the passed parameter
   followed by "TRUE" or "FALSE" depending on the value of the
   (integer) parameter.  */
char *make_bool_text(Component *c, Param *p)
{
  static char b[60];

  sprintf(b, "%s : %s", p->p_name, p->u.i ? "TRUE" : "FALSE");
  return(b);
}

char *make_short_bool_text(Component *c, Param *p)
{
  return(p->u.i ? "TRUE" : "FALSE");
}


/* Functions to maintain a parameter which is a string. */
char *make_string_text(Component *c, Param *p)
{
    static char b[512];

    sprintf(b, "%s: %.*s", p->p_name, sizeof(b)-sizeof(p->p_name)-3,
	    p->u.p ? p->u.p : "");
    return(b);
}


char *make_short_string_text(Component *c, Param *p)
{
  static char b[512];

  sprintf(b, "'%.*s'", sizeof(b)-1, p->u.p ? p->u.p : "");
  return(b);
}


double bool_calc(Component *c, Param *p)
{
  /* Very simple:  Return 1.0 if true, 0.0 if not.  Really should be
     something that changes colors rather than a meter, but we don't have
     anything besides meters yet. */

  return(p->u.i ? 1.0 : 0.0);
}

double color_calc(Component *c, Param *p)
{
  /* The value of the param is an integer number of a color.
     Just convert it to a double & return it. */
  return((double)p->u.i);
}


/* SHIV : CISE : unsure */ 
double pktq_calc(Component *c, Param *p)
{
  return (double)((queue *)p->u.p)->q_len;
}

/* SHIV : CISE : unsure */
char * pktq_text(Component *c, Param *p)
{
  int qlen = ((queue *)p->u.p)->q_len;
  static char b[80];

  sprintf(b, "%s has %d pkts", p->p_name, qlen);
  return(b);
}

/* SHIV : CISE : unsure */ 
char * pktq_short_text(Component *c, Param *p)
{
  int qlen = ((queue *)p->u.p)->q_len;
  static char b[20];

  sprintf(b, "%d", qlen);
  return(b);
}


/* Generic queue routines (That don't say "packet" in their output). */

double q_calc(Component *c, Param *p)
{
  int qlen = ((queue *)p->u.p)->q_len;

  /* Do the right thing--don't do any scaling here. */
  return((double)qlen);
}

char *q_text(Component *c, Param *p)
{
  int qlen = ((queue *)p->u.p)->q_len;
  static char b[80];

  sprintf(b, "%s length %d", p->p_name, qlen);
  return(b);
}

char *q_short_text(Component *c, Param *p)
{
  int qlen = ((queue *)p->u.p)->q_len;
  static char b[20];

  sprintf(b, "%d", qlen);
  return(b);
}


/* Routine to do input for a param.  Just converts string to int. */
int param_input_int(Param *p, char *s, Component *c)
{
  /* Minimal error check:  make sure that first non-whitespace char
     is a digit. */
  for (; isspace(*s); s++)
    ;
  if (isdigit(*s) || *s == '-')  {
    p->u.i = atoi(s);
    return(TRUE);
  }
  return(FALSE);
}

int param_input_double(Param *p, char *s, Component *c)
{
  /* Minimal error check:  make sure that first non-whitespace char
     is a digit. */
  for (; isspace(*s); s++)
    ;
  if (isdigit(*s) || *s == '-' || *s == '.')  {
    /* p->u.d = atof(s); */
    sscanf(s, "%lf", &p->u.d);
    return(TRUE);
  }
  return(FALSE);
}

int param_input_bool(Param *p, char *s, Component *c)
{
  char buf[256];

  /* Skip whitespace */
  for (; isspace(*s); s++)
    ;
  /* Copy the passed string into the buffer 'cause it's going to be
     changed. */
  strcpy(buf, s);

  ucase(buf);
  if (strcmp("TRUE", buf) == 0 || atoi(buf) != 0 || strcmp("T", buf) == 0)
    p->u.i = TRUE;
  else
    p->u.i = FALSE;

  return(TRUE);
}

int param_input_string(Param *p, char *s, Component *c)
{
    if (p->u.p)
	free (p->u.p);

    p->u.p = sim_malloc (strlen (s)+1);
    strcpy (p->u.p, s);
    return TRUE;
}


/*******************************************************
 * Helper routines to implement an "enum" parameter type.
 * The parameter value is an integer from a named set of integers.
 * The helper routines below are used to construct the real
 * parameter I/O routines as follows:
 * For example, assume that we want to the store the values
 * "apple", "orange", and "bananna" in the parameter.
 * Then the "fruit" I/O routines would be defined:
 * char *fruits[] = {"appple", "orange", "bananna"};
 # #define NFRUITS 3 
 * param_input_fruit (p, s, c)
 * {
 *	return param_input_enum_helper (p, s, c, fruits, NFRUITS);
 * }
 *
 * char *make_fruit_text (c, p)
 * {
 *	return make_enum_text_helper (c, p, fruits, NFRUITS);
 * }
 * etc.
 * The value stored in the parameter will be the index
 * of the corresponding string in the fruits array.
 * param_input_enum_helper() considers s to match the first entry for which
 * is is a prefix.
 */

param_input_enum_helper (Param *p, char *s, Component *c, 
			 char *string_list[], int n)
{
  int i;

  for (i = 0; i < n; i++)  {
    char *match;
    match = strstr (string_list[i], s);
    /* s must be a prefix of the entry (not just a substring)  */
    if (match && match == string_list[i])  {
      p->u.i = i;
      return TRUE;
    }
  }

  return FALSE;
}

char *make_enum_text_helper (Component *c, Param *p,
			     char *string_list[], int n)
{
  static char b[140];
  if (p->u.i >= n || p->u.i < 0)
    sprintf (b, "%s: %d (INVALID)", c->co_name, p->u.i);
  else
    sprintf (b, "%s: %.99s", p->p_name, string_list[p->u.i]);

  return b;
}

char *make_short_enum_text_helper (Component *c, Param *p,
			     char *string_list[], int n)
{
  static char b[100];
  if (p->u.i >= n || p->u.i < 0)
    sprintf (b, "%d (INVALID)", p->u.i);
  else
    sprintf (b, "%.99s", string_list[p->u.i]);

  return b;
}


/*******************************************************
  Routines to replace calloc(), malloc(), and realloc().  They
  are guaranteed to either return a valid pointer or to not return.
  If the allocation fails, they cause a core dump. */

char *sim_malloc(unsigned size)
{
  register char *p = (char *)malloc(size);

  if (!p)
    panic("No memory\n");

  return(p);
}

char *sim_calloc(unsigned nelem, unsigned elsize)
{
  register char *p = (char *)calloc(nelem, elsize);
 
  if (!p)
    panic("No memory\n");

  return(p);
}

char *sim_realloc(char *ptr, unsigned size)
{
  register char *p = (char *)realloc(ptr, size);

  if (!p)
    panic("No memory\n");

  return(p);
}


/* Print a message and cause a core dump */
void panic(char *s)
{
  fputs(s, stderr);
  fflush(stderr);
  abort();
}



/**************************************************
  Miscellaneous useful functions.
*/


/*  Routine to find a component in a list of neighbors.  */
Neighbor *find_neighbor(list *l, Component *c)
{
  register Neighbor *n;

  for (n = (Neighbor *)l->l_head; n; n = n->n_next)
    if (n->n_c == c)
      break;

  return(n);
}


/* Upper caseify a string */
int ucase(register char *s)
{
  while (*s)  {
    if (islower(*s))
      *s = toupper(*s);
    s++;
  }
  return 0;
}



/**** Routine to initialize a parameter.
  Pass it all the fields and a component, allocates a parameter,
  stuffs the fields into it, and adds it to the queue.  The address
  of the new parameter is returned. */

Param *param_init(Component *c, char *name, PFD calc_val, 
		  PFP make_text, PFP make_short_text, PFI input,
		  int display_type, int flags, double scale)
{
  register Param *p = (Param *)sim_calloc(1, sizeof(Param));

  if (!p)
    return((Param *)NULL);

  strcpy(p->p_name, name);
  p->p_calc_val = calc_val;
  p->p_make_text = make_text;
  p->p_make_short_text = make_short_text;
  p->p_input = input;
  p->p_display_type = display_type;
  p->p_flags = flags;
  p->p_scale = scale;

  qe_addt(c->co_params, p);
  return(p);
}


/*******************************************************
  Functions to perform housekeeping that all components must do.
*/


/******** Things that all components do when they are deleted. */
comp_delete(Component *c)
{
    Neighbor *n;

    /* UNEIGHBOR self from all neighbors */
    for (n = (Neighbor *)c->co_neighbors->l_head; n; n = n->n_next)
	if (!((*(n->n_c->co_action))(c, n->n_c, EV_UNEIGHBOR,NULL,(char *)c))) {
	    dbg_write(DBG_ERR, c,
		      "couldn't remove self from a neighbor while deleting");
	}
  
    /* Wipe out the neighbor list */
    lq_delete(c->co_neighbors);
    /* Destroy parameter storage */
    lq_delete(c->co_params);
    /* Wipe out this component */
    free((char *)c);
}

/*********
  Add a neighbor to a component.
  Takes care of:
	Checking for max number of neighbors.
	Checking for duplicate neighbors.
	Only correct classes of neighbors.
  Any other types of error checking should be done *before* calling this,
  since this routine actually allocates memory and changes the
  component structure.
  This function returns a pointer to the new neighbor structure.
  The caller must initialize any parameters and packet queues in
  that neighbor structure.
*/

Neighbor *add_neighbor(register Component *c, register Component *neighc,
		       int max_num_neighbors, int num_classes, ...)
    /*    c:                  Comp to add neighbor to  */
    /*    neighc:             New neighbor  */
    /*    max_num_neighbors:  Max number neighbors allowed (0=infinite) */
    /*    num_classes:        How many classes follow  */
{
    va_list  vp;
    register Neighbor *n;
    int      notlegal;

    /* Check that have less than the max number of neighbors */
    /* max_num_neighbors <= 0 means infinite number is permissible */
    if (max_num_neighbors > 0)
	if (c->co_neighbors->l_len >= max_num_neighbors)  {
	    dbg_write(DBG_ERR, c,
		      "cannot attach more than %d neighbors", max_num_neighbors);
	    return((Neighbor *)NULL);
	}
    
    /* Check that it is not already a neighbor */
    /* we are going to allow adding a link to a leaky_bucket component twice */
    if (find_neighbor(c->co_neighbors, neighc))  {
	dbg_write(DBG_ERR, c,
		  "attempt to attach '%s' twice as neighbor", neighc->co_name);
	return((Neighbor *)NULL);
    }

    if (num_classes > 0)  {
	/* Final check--may sure the new neighbor is one of the
	   legal classes */
	va_start(vp,num_classes);
	notlegal = TRUE;
	while (num_classes--)
	    if (neighc->co_class == va_arg(vp, int))  {
		notlegal = FALSE;
		break;
	    }
	va_end(vp);
    }

    if (notlegal)  {
	dbg_write(DBG_ERR, c, "attempt to attach '%s'--not a legal neighbor",
		  neighc->co_name);
	return((Neighbor *)NULL);
    }

    /** Finally.... can really do it. */
    n = (Neighbor *)sim_calloc(1, sizeof(Neighbor));
    n->n_c = neighc;
    le_addt(c->co_neighbors, n);

    dbg_write(DBG_INFO, c, "added '%s' as neighbor", neighc->co_name);

    return(n);
}


/**********
  Generic routine to remove a neighbor from a component.  Is passed a
  pointer to the neighbor component.  
*/

int remove_neighbor(register Component *c, register Component *neighc)
{
    register Neighbor *n;

    n = find_neighbor(c->co_neighbors, neighc);
    if (n)  {
	le_del(c->co_neighbors, n);
	if (n->n_pq)
	    lq_delete(n->n_pq);
	if (n->n_p)  {
	    qe_del(c->co_params, n->n_p);
	    free(n->n_p);
	}
	if (n->n_pp)  {
	    qe_del(c->co_params, n->n_pp);
	    free(n->n_pp);
	}
	if (n->n_data) 
	    free (n->n_data);
	free(n);
	dbg_write(DBG_INFO, c, "removed '%s' from neighbor list", 
		  neighc->co_name);
	return(TRUE);
    }

    dbg_write(DBG_ERR, c, "tried to remove nonexistant neighbor '%s'", 
	      neighc->co_name);

    return(FALSE);
}


/* usleep() is a BSDism. If not BSD, define our own */
 
void usleep(unsigned long usec)
{
	/*
    struct timeval timeout;
    timeout.tv_sec  = usec/1000000;
    timeout.tv_usec = usec%1000000;
    select(0, NULL, NULL, NULL, &timeout);
	 */
	/* Sleep(usec); */ 
}
 

/* In Win95, there is no bcopy, use memmove instead */
 void bcopy(char *b1, char *b2, int length)
{
	memmove(b2, b1, length);
}



/*** Return a random number r, s.t. a <= r <= b */

double random_range(register double a, register double b)
{
  register double diff = b - a;

  return ((((double)Cise_random() / (double)MAX_RANDOM_NUMBER) * diff) + a);
}



#if (defined(hppa) || defined(UNICOS))

/*

     Rint returns the integer (represented as a double precision
     number) nearest x in the direction of the prevailing round-
     ing mode.

     On a VAX, rint(x) is equivalent to adding half to the magni-
     tude and then rounding towards zero.

     In the default rounding mode, to nearest, on a machine that
     conforms to IEEE 754, rint(x) is the integer nearest x with
     the additional stipulation that if |rint(x)-x|=1/2 then
     rint(x) is even.  Other rounding modes can make rint act
     like floor, or like ceil, or round towards zero.

     Another way to obtain an integer near x is to declare (in C)
          double x;     int k;    k = x;
     Most C compilers round x towards 0 to get the integer k, but
     some do otherwise.  If in doubt, use floor, ceil, or rint
     first, whichever you intend.  Also note that, if x is larger
     than k can accommodate, the value of k and the presence or
     absence of an integer overflow are hard to predict.

 */

#include	<math.h>

double		rint (double x)

{
	double		flr;

	flr = floor (x);
	return (((x - flr) >= (double) 0.5) ? ceil (x) : flr);
}


     

#endif 
